/*=========================================================================

  Program:   Visualization Toolkit Bic Extension
  Module:    $RCSfile: vtkMINCReader.cxx,v $
  Language:  C++
  Date:      $Date: 2009-04-04 20:59:39 $
  Version:   $Revision: 1.9 $

  Copyright (c) 2002-2008 IPL, BIC, MNI, McGill, Simon Drouin, Eric Benoit
  All rights reserved.

=========================================================================*/
#include "vtkMINCReader.h"
#include "vtkImageData.h"
#include "vtkPointData.h"
#include "vtkTransform.h"
#include "vtkMatrix4x4.h"


vtkCxxRevisionMacro(vtkMINCReader, "$Revision: 1.9 $");


vtkMINCReader * vtkMINCReader::New()
{
    return new vtkMINCReader();
}


void vtkMINCReader::SetIcv( vtkMINCIcv * icv )
{
    if( this->Icv != NULL )
        this->Icv->Delete();

    this->Icv = icv;
    this->Icv->Register( this );
}


//-------------------------------------------------------------------------------
vtkMINCReader::vtkMINCReader()
{
    this->Icv = NULL;
    this->Transform = vtkTransform::New();
    this->Transform->Identity();

    this->MincFileId = -1;
    this->ImageVarId = -1;
    this->ImageDataType = VTK_UNSIGNED_CHAR;
    this->NumberOfScalarComponents = 1;
    for( int i = 0; i < 4; i++ )
    {
        this->DimIndex[i] = -1;
        this->DimensionLength[i] = 1;
        this->DimensionNames[1][0] = '\0';
        this->Step[i] = 1;
        this->Origin[i] = 0;
        for( int j = 0; j < 3; j++ )
        {
            if( i == j )
                this->DirectionCosines[i][j] = 1.0;
            else
                this->DirectionCosines[i][j] = 0.0;
        }
    }

    this->DimensionOrder[0] = 0;
    this->DimensionOrder[1] = 1;
    this->DimensionOrder[2] = 2;
    ncopts = NC_VERBOSE; // | NC_FATAL;
    this->UseTransformForStartAndStep = 0;
    this->ReaderStatus = MINC_OK;
}


//-------------------------------------------------------------------------------
vtkMINCReader::~vtkMINCReader()
{
   if( this->Icv != NULL )
   {
       this->Icv->Delete();
   }
   if( this->MincFileId != MI_ERROR )
       miclose( this->MincFileId );

   this->Transform->Delete();
}

//-------------------------------------------------------------------------------
int vtkMINCReader::CanReadFile( const char * fname )
{
    char name[1024];       // filenames shouldn't be longer than that
    strcpy( name, fname ); // have to do that because miopen doesn't take const char *
    this->MincFileId = miopen( name, NC_NOWRITE );
    if ( this->MincFileId == -1 )
    {
        return 0;
    }
    miclose( this->MincFileId );
    this->MincFileId = MI_ERROR;
    return 3;
}

//----------------------------------------------------------------------------
// This templated function reorders pixels for any type of data.
template < class type >
void vtkReorderVoxels(   vtkMINCReader * self,
                         type * inptr,
                         type * outptr,
                         int outputStep[3],
                         int outputExtent[6],
                         int numberOfScalarComponent )
{
    type * inIt[ 3 ];
    inIt[ 0 ] = inptr;
    inIt[ 1 ] = inptr;
    inIt[ 2 ] = inptr;
    type * outIt = outptr;
    int numberOfSlices = outputExtent[5] - outputExtent[4] + 1;
    double percent = 0;

    self->SetProgressText("Reordering voxels");

    for( int i = outputExtent[ 4 ] ; i <= outputExtent[ 5 ]; i++ )
    {
        inIt[ 1 ] = inIt[ 2 ];
        for( int j = outputExtent[ 2 ]; j <= outputExtent[ 3 ] ; j++ )
        {
            inIt[ 0 ] = inIt[ 1 ];
            for( int k = outputExtent[ 0 ]; k <= outputExtent[ 1 ]; k++ )
            {
                for( int l = 0; l < numberOfScalarComponent; l++ )
                {
                    outIt[l] = inIt[0][l];
                }
                outIt += numberOfScalarComponent;
                inIt[ 0 ] += outputStep[ 0 ];
            }
            inIt[ 1 ] += outputStep[ 1 ];
        }
        inIt[ 2 ] += outputStep[ 2 ];
        percent = (double)( i - outputExtent[4] ) / numberOfSlices;
        self->UpdateProgress( percent );
    }
}


//-------------------------------------------------------------------------------
// This function reads a data from a file.  The datas extent/axes
// are assumed to be the same as the file extent/order.
void vtkMINCReader::ExecuteData(vtkDataObject *output)
{
    if (this->ReaderStatus != MINC_OK)
        return;

    vtkImageData * imageData = this->AllocateOutputData(output);
    
    //=============================================
    // Open the minc file
    //=============================================
    if ( this->InternalFileName == NULL )
    {
        vtkErrorMacro(<< "Either a FileName or FilePrefix must be specified.");
        this->ReaderStatus = MINC_CORRUPTED;
        return;
    }

    if( this->MincFileId != MI_ERROR )
    {
        miclose( this->MincFileId );
    }

    this->MincFileId = miopen( this->InternalFileName, NC_NOWRITE );
    if ( this->MincFileId == -1 )
    {
        vtkErrorMacro(<<"Could not open minc file.");
        this->ReaderStatus = MINC_CORRUPTED;
        return;
    }

    this->ComputeDataIncrements();
    imageData->GetPointData()->GetScalars()->SetName("MINCImage");

    //=============================================
    // Compute the starts and lengths to read all
    // voxels.
    //=============================================
    int extent[6];
    imageData->GetExtent( extent );

    long start[4] = { 0, 0, 0, 0 };
    long dimLength[4] = { 1, 1, 1, 1 };

    this->SetProgressText("Reading voxels");

    int i;
    for( i = 0; i < 3; i++ )
    {
        if( this->DimIndex[i] != -1 )
        {
            start[ this->DimIndex[i] ] = extent[ 2 * i ];
            dimLength[ this->DimIndex[i] ] = extent[ 2 * i + 1 ] - extent[ 2 * i ] + 1;
        }
    }
    if( this->DimIndex[3] != -1 )
    {
        dimLength[ this->DimIndex[3] ] = this->NumberOfScalarComponents;
    }


    //=============================================
    // Read voxels
    //=============================================
    vtkImageData * tempImage = vtkImageData::New();
    tempImage->SetNumberOfScalarComponents( this->NumberOfScalarComponents );
    tempImage->SetScalarType( this->GetDataScalarType() );
    tempImage->SetExtent( extent );
    tempImage->AllocateScalars();
    void * outPtr = tempImage->GetScalarPointer();
    if( this->Icv )
    {
        int id = this->Icv->GetIcvId();
        miicv_get( id, start, dimLength, outPtr );
    }
    else
    {
        nc_type mincDataType = NC_SHORT;
        int isSigned = 1;
        char sign[9] = MI_SIGNED;
        vtkMINCIcv::VtkDataTypeToMincDataType( this->GetDataScalarType(), mincDataType, isSigned );
        strcpy( sign, isSigned ? MI_SIGNED : MI_UNSIGNED );
        if( mivarget( this->MincFileId,
                      this->ImageVarId,
                      start, dimLength,
                      mincDataType,
                      sign,
                      outPtr )  == MI_ERROR )
        {
            vtkErrorMacro(<<"Error getting image voxels." );
            this->ReaderStatus = MINC_CORRUPTED;
            return;
        }
    }


    //=============================================
    // Reorder voxels
    //=============================================
    int inputDim[3];
    int inputStep[3];
    inputStep[0] = this->NumberOfScalarComponents;
    inputStep[1] = this->NumberOfScalarComponents;
    inputStep[2] = this->NumberOfScalarComponents;



    for( i = 0; i < 3; i++ )
    {
        inputDim[ i ] = dimLength[ this->NumberOfSpacialDimensions - i - 1 ];
        for( int j = 0; j < i; j++ )
        {
            inputStep[ i ] *= inputDim[ j ];
        }
    }

    int outputStep[3];
    for( i = 0; i < 3; i++ )
    {
        outputStep[ this->DimensionOrder[ i ] ] = inputStep[ i ];
    }

    switch ( imageData->GetScalarType() )
    {
    vtkTemplateMacro6(vtkReorderVoxels,
                      this,
                      static_cast<VTK_TT *>(tempImage->GetScalarPointer()),
                      static_cast<VTK_TT *>(imageData->GetScalarPointer()),
                      outputStep, extent, this->NumberOfScalarComponents );
    default:
      vtkErrorMacro(<< "Execute: Unknown ScalarType");
      this->ReaderStatus = MINC_CORRUPTED;
      return;
    }

    tempImage->Delete();
}


//------------------------------------------------------------------------------
void vtkMINCReader::ExecuteInformation()
{
    this->ComputeInternalFileName(this->DataExtent[4]);
    if ( this->InternalFileName == NULL )
    {
        vtkErrorMacro(<<"No valid minc file name.");
        this->ReaderStatus = MINC_CORRUPTED;
        return;
    }

    if( this->MincFileId != MI_ERROR )
    {
        miclose( this->MincFileId );
    }

    //=============================================
    // Open the minc file
    //=============================================
    this->MincFileId = miopen( this->InternalFileName, NC_NOWRITE );
    if ( this->MincFileId == -1 )
    {
        vtkErrorMacro(<<"Could not open minc file.");
        this->ReaderStatus = MINC_CORRUPTED;
        this->DataExtent[0] = 0;
        this->DataExtent[1] = 0;
        this->DataExtent[2] = 0;
        this->DataExtent[3] = 0;
        this->DataExtent[4] = 0;
        this->DataExtent[5] = 0;
        this->SetNumberOfScalarComponents(1);
        this->vtkImageReader2::ExecuteInformation();
        return;
    }

    //=============================================
    // Get the id of the image variable
    //=============================================
    if( nc_inq_varid ( this->MincFileId, MIimage, &( this->ImageVarId ) ) == MI_ERROR )
    {
        vtkErrorMacro(<<"Could not get the id of the image variable");
        this->ReaderStatus = MINC_CORRUPTED;
        return;
    }

    //=============================================
    // Create an icv and attach it to the image
    // variable.
    //=============================================
    if( this->Icv != NULL )
    {
        Icv->SetupIcv();
        miicv_attach( Icv->GetIcvId(), this->MincFileId, this->ImageVarId );
    }


    //=============================================
    // Get the number of dimensions
    //=============================================
    int numberOfDimensions;
    if( ncinquire( this->MincFileId, &numberOfDimensions, NULL, NULL, NULL) == MI_ERROR )
    {
        vtkErrorMacro(<<"Error getting the number of dimensions");
        this->ReaderStatus = MINC_CORRUPTED;
        return;
    }

    if( numberOfDimensions > MAX_NUMBER_OF_DIMENSIONS )
    {
        vtkErrorMacro(<<"The maximum number of dimensions supported by this reader is " << MAX_NUMBER_OF_DIMENSIONS << " (including vector dimension)");
        this->ReaderStatus = MINC_CORRUPTED;
        return;
    }

    //=============================================
    // Get the attributes of every dimension
    //=============================================
    int dimVarId = MI_ERROR;
    int i;
    for( i = 0 ; i < numberOfDimensions; i++ )
    {
        if( ncdiminq( this->MincFileId, i, this->DimensionNames[i], NULL ) == MI_ERROR )
        {
            vtkErrorMacro(<<"Error getting dimension " << i << " attributes." << endl);
            this->ReaderStatus = MINC_CORRUPTED;
            return;
        }
        else
        {
            if( !strcmp( MIxspace, this->DimensionNames[i] ) )
            {
                this->DimIndex[0] = i;
            }
            else if( !strcmp( MIyspace, this->DimensionNames[i] ) )
            {
                this->DimIndex[1] = i;
            }
            else if( !strcmp( MIzspace, this->DimensionNames[i] ) )
            {
                this->DimIndex[2] = i;
            }
            else if( !strcmp( MIvector_dimension, this->DimensionNames[i] ) )
            {
                this->DimIndex[3] = i;
            }
        }
        if( ncdiminq( this->MincFileId, i, NULL, &(this->DimensionLength[i]) ) == MI_ERROR )
        {
            vtkErrorMacro(<<"Error getting dimension " << i << " length.");
            this->ReaderStatus = MINC_CORRUPTED;
            return;
        }

        dimVarId = ncvarid( this->MincFileId, this->DimensionNames[i] );
        if( dimVarId != MI_ERROR )
        {
            miattget1( this->MincFileId, dimVarId, MIstep, NC_DOUBLE, &(this->Step[i]) );
            miattget1( this->MincFileId, dimVarId, MIstart, NC_DOUBLE, &(this->Origin[i]) );
            miattget( this->MincFileId, dimVarId, MIdirection_cosines, NC_DOUBLE, 4, &(this->DirectionCosines[i]), NULL );
            if( this->Step[i] < 0 )
            {
                vtkWarningMacro(<< " At least one of the step size is negative in the image. This is not well supported by some display classes. To changed that: mincreshape +direction <input_file> <output_file>" << endl );
                this->ReaderStatus = MINC_NEGATIVE_STEP;
            }
        }
    }

    this->NumberOfSpacialDimensions = numberOfDimensions;
    if( this->DimIndex[3] != -1 )
    {
        this->NumberOfSpacialDimensions--;
        this->NumberOfScalarComponents = this->DimensionLength[ this->DimIndex[3] ];
        if( this->DimIndex[3] != this->NumberOfSpacialDimensions )
        {
            vtkErrorMacro(<<"Error: this reader does not support minc files where vector dimension is not the last");
            this->ReaderStatus = MINC_CORRUPTED;
            return;
        }
    }

    //=============================================
    // compute the dimension order
    //=============================================
    for( i = 0; i < 3; i++ )
    {
        if( this->DimIndex[i] != -1 )
        {
            this->DimensionOrder[ this->NumberOfSpacialDimensions - this->DimIndex[i] - 1 ] = i;
        }
    }

    //=============================================
    // Compute a transform from start, step and
    // direction cosines.
    //=============================================
    double translation[3] = { 0.0, 0.0, 0.0 };
    double scale[3] = { 1.0, 1.0, 1.0 };
    vtkMatrix4x4 * rotation = vtkMatrix4x4::New();

    for( i = 0; i < 3; i++ )
    {
        if( this->DimIndex[i] != -1 )
        {
            translation[i] = this->Origin[ this->DimIndex[i] ];
            scale[i] = this->Step[ this->DimIndex[i] ];
            for( int j = 0; j < 3; j++ )
            {
                rotation->SetElement( j, i, this->DirectionCosines[this->DimIndex[i]][j] );
            }
        }
    }

    this->Transform->Identity();
    this->Transform->Concatenate( rotation );
    if( this->UseTransformForStartAndStep )
    {
        this->Transform->Translate( translation );
        this->Transform->Scale( scale );
    }
    else
    {
        this->SetDataSpacing( scale[0], scale[1], scale[2] );
        this->SetDataOrigin( translation[0], translation[1], translation[2] );
    }
    rotation->Delete();


    //=============================================
    // Get the image variable data type
    //=============================================
    if( !this->Icv )
    {
        nc_type imageDataType;
        int imageDataTypeIsSigned;
        if( miget_datatype( this->MincFileId, this->ImageVarId, &imageDataType, &imageDataTypeIsSigned ) == MI_ERROR )
        {
            vtkErrorMacro(<<"Could not get image data type");
            this->ReaderStatus = MINC_CORRUPTED;
            return;
        }
        this->ImageDataType = vtkMINCIcv::MincDataTypeToVtkDataType( imageDataType, imageDataTypeIsSigned );
    }
    else
    {
        this->ImageDataType = Icv->GetDataType();
    }


    //=============================================
    // Set image output extent
    //=============================================
    this->DataExtent[0] = 0;
    this->DataExtent[1] = 0;
    this->DataExtent[2] = 0;
    this->DataExtent[3] = 0;
    this->DataExtent[4] = 0;
    this->DataExtent[5] = 0;
    for( i = 0; i < 3; i++ )
    {
        if( this->DimIndex[i] != -1 )
        {
            this->DataExtent[ 2 * i + 1 ] = this->DimensionLength[ this->DimIndex[i] ] - 1;
        }
    }

    //=============================================
    // Set the number of scalar components
    //=============================================
    this->SetDataScalarType( this->ImageDataType );
    this->SetNumberOfScalarComponents( this->NumberOfScalarComponents );

    //=============================================
    // Let parent class do the rest.
    //=============================================
    this->vtkImageReader2::ExecuteInformation();

    miclose( this->MincFileId );
    this->MincFileId = MI_ERROR;
}


//----------------------------------------------------------------------------
void vtkMINCReader::PrintSelf(ostream& os, vtkIndent indent)
{
    this->Superclass::PrintSelf(os,indent);
}
